@@ -4,7 +4,7 @@ When the source code of a module has a change in fine-grained incremental mode,
 we build a new AST from the updated source. However, other parts of the program
 may have direct references to parts of the old AST (namely, those nodes exposed
 in the module symbol table). The merge operation changes the identities of new
-AST nodes that have a correspondance in the old AST to the old ones so that
+AST nodes that have a correspondence in the old AST to the old ones so that
 existing cross-references in other modules will continue to point to the correct
 nodes. Also internal cross-references within the new AST are replaced. AST nodes
 that aren't externally visible will get new, distinct object identities. This
@@ -15,7 +15,7 @@ external references (which would be slow and fragile) or always perform
 translation when looking up references (which would be hard to retrofit).
 
 The AST merge operation is performed after semantic analysis. Semantic
-analysis has to deal with potentionally multiple aliases to certain AST
+analysis has to deal with potentially multiple aliases to certain AST
 nodes (in particular, MypyFile nodes). Type checking assumes that we
 don't have multiple variants of a single AST node visible to the type
 checker.
@@ -28,7 +28,7 @@ Discussion of some notable special cases:
 
 * If a function is replaced with another function with an identical signature,
   call sites continue to point to the same object (by identity) and don't need
-  to be reprocessed. Similary, if a class is replaced with a class that is
+  to be reprocessed. Similarly, if a class is replaced with a class that is
   sufficiently similar (MRO preserved, etc.), class references don't need any
   processing. A typical incremental update to a file only changes a few
   externally visible things in a module, and this means that often only few
@@ -48,17 +48,20 @@ See the main entry point merge_asts for more details.
 from typing import Dict, List, cast, TypeVar, Optional
 
 from mypy.nodes import (
-    Node, MypyFile, SymbolTable, Block, AssignmentStmt, NameExpr, MemberExpr, RefExpr, TypeInfo,
+    MypyFile, SymbolTable, Block, AssignmentStmt, NameExpr, MemberExpr, RefExpr, TypeInfo,
     FuncDef, ClassDef, NamedTupleExpr, SymbolNode, Var, Statement, SuperExpr, NewTypeExpr,
-    OverloadedFuncDef, LambdaExpr, TypedDictExpr, EnumCallExpr, MDEF
+    OverloadedFuncDef, LambdaExpr, TypedDictExpr, EnumCallExpr, FuncBase, TypeAliasExpr, CallExpr,
+    CastExpr,
+    MDEF
 )
 from mypy.traverser import TraverserVisitor
 from mypy.types import (
-    Type, TypeVisitor, Instance, AnyType, NoneTyp, CallableType, DeletedType, PartialType,
+    Type, SyntheticTypeVisitor, Instance, AnyType, NoneTyp, CallableType, DeletedType, PartialType,
     TupleType, TypeType, TypeVarType, TypedDictType, UnboundType, UninhabitedType, UnionType,
-    Overloaded, TypeVarDef
+    Overloaded, TypeVarDef, TypeList, CallableArgument, EllipsisType, StarType
 )
-from mypy.util import get_prefix
+from mypy.util import get_prefix, replace_object_state
+from mypy.typestate import TypeState
 
 
 def merge_asts(old: MypyFile, old_symbols: SymbolTable,
@@ -156,34 +159,41 @@ class NodeReplaceVisitor(TraverserVisitor):
 
     def visit_func_def(self, node: FuncDef) -> None:
         node = self.fixup(node)
-        if node.type:
-            self.fixup_type(node.type)
-        if node.info:
-            node.info = self.fixup(node.info)
+        self.process_base_func(node)
         super().visit_func_def(node)
 
     def visit_overloaded_func_def(self, node: OverloadedFuncDef) -> None:
-        if node.info:
-            node.info = self.fixup(node.info)
+        self.process_base_func(node)
         super().visit_overloaded_func_def(node)
 
     def visit_class_def(self, node: ClassDef) -> None:
         # TODO additional things?
+        node.info = self.fixup_and_reset_typeinfo(node.info)
         node.defs.body = self.replace_statements(node.defs.body)
-        node.info = self.fixup(node.info)
+        info = node.info
         for tv in node.type_vars:
             self.process_type_var_def(tv)
-        self.process_type_info(node.info)
+        if info:
+            if info.is_named_tuple:
+                self.process_synthetic_type_info(info)
+            else:
+                self.process_type_info(info)
         super().visit_class_def(node)
 
+    def process_base_func(self, node: FuncBase) -> None:
+        self.fixup_type(node.type)
+        node.info = self.fixup(node.info)
+        if node.unanalyzed_type:
+            # Unanalyzed types can have AST node references
+            self.fixup_type(node.unanalyzed_type)
+
     def process_type_var_def(self, tv: TypeVarDef) -> None:
         for value in tv.values:
             self.fixup_type(value)
         self.fixup_type(tv.upper_bound)
 
     def visit_assignment_stmt(self, node: AssignmentStmt) -> None:
-        if node.type:
-            self.fixup_type(node.type)
+        self.fixup_type(node.type)
         super().visit_assignment_stmt(node)
 
     # Expressions
@@ -200,46 +210,59 @@ class NodeReplaceVisitor(TraverserVisitor):
     def visit_ref_expr(self, node: RefExpr) -> None:
         if node.node is not None:
             node.node = self.fixup(node.node)
+            if isinstance(node.node, Var):
+                # The Var node may be an orphan and won't otherwise be processed.
+                fixup_var(node.node, self.replacements)
 
     def visit_namedtuple_expr(self, node: NamedTupleExpr) -> None:
         super().visit_namedtuple_expr(node)
-        node.info = self.fixup(node.info)
-        self.process_type_info(node.info)
+        node.info = self.fixup_and_reset_typeinfo(node.info)
+        self.process_synthetic_type_info(node.info)
+
+    def visit_cast_expr(self, node: CastExpr) -> None:
+        super().visit_cast_expr(node)
+        self.fixup_type(node.type)
 
     def visit_super_expr(self, node: SuperExpr) -> None:
         super().visit_super_expr(node)
         if node.info is not None:
             node.info = self.fixup(node.info)
 
+    def visit_call_expr(self, node: CallExpr) -> None:
+        super().visit_call_expr(node)
+        if isinstance(node.analyzed, SymbolNode):
+            node.analyzed = self.fixup(node.analyzed)
+
     def visit_newtype_expr(self, node: NewTypeExpr) -> None:
         if node.info:
-            node.info = self.fixup(node.info)
-            self.process_type_info(node.info)
-        if node.old_type:
-            self.fixup_type(node.old_type)
+            node.info = self.fixup_and_reset_typeinfo(node.info)
+            self.process_synthetic_type_info(node.info)
+        self.fixup_type(node.old_type)
         super().visit_newtype_expr(node)
 
     def visit_lambda_expr(self, node: LambdaExpr) -> None:
-        if node.info:
-            node.info = self.fixup(node.info)
+        node.info = self.fixup(node.info)
         super().visit_lambda_expr(node)
 
     def visit_typeddict_expr(self, node: TypedDictExpr) -> None:
-        node.info = self.fixup(node.info)
         super().visit_typeddict_expr(node)
+        node.info = self.fixup_and_reset_typeinfo(node.info)
+        self.process_synthetic_type_info(node.info)
 
     def visit_enum_call_expr(self, node: EnumCallExpr) -> None:
-        node.info = self.fixup(node.info)
-        self.process_type_info(node.info)
+        node.info = self.fixup_and_reset_typeinfo(node.info)
+        self.process_synthetic_type_info(node.info)
         super().visit_enum_call_expr(node)
 
+    def visit_type_alias_expr(self, node: TypeAliasExpr) -> None:
+        self.fixup_type(node.type)
+        super().visit_type_alias_expr(node)
+
     # Others
 
     def visit_var(self, node: Var) -> None:
-        if node.info:
-            node.info = self.fixup(node.info)
-        if node.type:
-            self.fixup_type(node.type)
+        node.info = self.fixup(node.info)
+        self.fixup_type(node.type)
         super().visit_var(node)
 
     # Helpers
@@ -247,29 +270,51 @@ class NodeReplaceVisitor(TraverserVisitor):
     def fixup(self, node: SN) -> SN:
         if node in self.replacements:
             new = self.replacements[node]
-            new.__dict__ = node.__dict__
+            replace_object_state(new, node)
             return cast(SN, new)
         return node
 
-    def fixup_type(self, typ: Type) -> None:
-        typ.accept(TypeReplaceVisitor(self.replacements))
+    def fixup_and_reset_typeinfo(self, node: TypeInfo) -> TypeInfo:
+        """Fix-up type info and reset subtype caches.
+
+        This needs to be called at least once per each merged TypeInfo, as otherwise we
+        may leak stale caches.
+        """
+        if node in self.replacements:
+            # The subclass relationships may change, so reset all caches relevant to the
+            # old MRO.
+            new = cast(TypeInfo, self.replacements[node])
+            TypeState.reset_all_subtype_caches_for(new)
+        return self.fixup(node)
+
+    def fixup_type(self, typ: Optional[Type]) -> None:
+        if typ is not None:
+            typ.accept(TypeReplaceVisitor(self.replacements))
 
     def process_type_info(self, info: Optional[TypeInfo]) -> None:
         if info is None:
             return
-        # TODO: Additional things:
-        # - declared_metaclass
-        # - metaclass_type
-        # - _promote
-        # - typeddict_type
-        # - replaced
+        self.fixup_type(info.declared_metaclass)
+        self.fixup_type(info.metaclass_type)
+        self.fixup_type(info._promote)
+        self.fixup_type(info.tuple_type)
+        self.fixup_type(info.typeddict_type)
+        info.defn.info = self.fixup(info)
+        info.replaced = self.fixup(info.replaced)
         replace_nodes_in_symbol_table(info.names, self.replacements)
         for i, item in enumerate(info.mro):
             info.mro[i] = self.fixup(info.mro[i])
         for i, base in enumerate(info.bases):
             self.fixup_type(info.bases[i])
-        if info.tuple_type:
-            self.fixup_type(info.tuple_type)
+
+    def process_synthetic_type_info(self, info: TypeInfo) -> None:
+        # Synthetic types (types not created using a class statement) don't
+        # have bodies in the AST so we need to iterate over their symbol
+        # tables separately, unlike normal classes.
+        self.process_type_info(info)
+        for name, node in info.names.items():
+            if node.node:
+                node.node.accept(self)
 
     def replace_statements(self, nodes: List[Statement]) -> List[Statement]:
         result = []
@@ -280,7 +325,7 @@ class NodeReplaceVisitor(TraverserVisitor):
         return result
 
 
-class TypeReplaceVisitor(TypeVisitor[None]):
+class TypeReplaceVisitor(SyntheticTypeVisitor[None]):
     """Similar to NodeReplaceVisitor, but for type objects."""
 
     def __init__(self, replacements: Dict[SymbolNode, SymbolNode]) -> None:
@@ -315,7 +360,9 @@ class TypeReplaceVisitor(TypeVisitor[None]):
     def visit_overloaded(self, t: Overloaded) -> None:
         for item in t.items():
             item.accept(self)
-        t.fallback.accept(self)
+        # Fallback can be None for overloaded types that haven't been semantically analyzed.
+        if t.fallback is not None:
+            t.fallback.accept(self)
 
     def visit_deleted_type(self, typ: DeletedType) -> None:
         pass
@@ -326,7 +373,9 @@ class TypeReplaceVisitor(TypeVisitor[None]):
     def visit_tuple_type(self, typ: TupleType) -> None:
         for item in typ.items:
             item.accept(self)
-        typ.fallback.accept(self)
+        # Fallback can be None for implicit tuple types that haven't been semantically analyzed.
+        if typ.fallback is not None:
+            typ.fallback.accept(self)
 
     def visit_type_type(self, typ: TypeType) -> None:
         typ.item.accept(self)
@@ -345,6 +394,19 @@ class TypeReplaceVisitor(TypeVisitor[None]):
         for arg in typ.args:
             arg.accept(self)
 
+    def visit_type_list(self, typ: TypeList) -> None:
+        for item in typ.items:
+            item.accept(self)
+
+    def visit_callable_argument(self, typ: CallableArgument) -> None:
+        typ.typ.accept(self)
+
+    def visit_ellipsis_type(self, typ: EllipsisType) -> None:
+        pass
+
+    def visit_star_type(self, typ: StarType) -> None:
+        typ.type.accept(self)
+
     def visit_uninhabited_type(self, typ: UninhabitedType) -> None:
         pass
 
@@ -369,17 +431,16 @@ def replace_nodes_in_symbol_table(symbols: SymbolTable,
         if node.node:
             if node.node in replacements:
                 new = replacements[node.node]
-                new.__dict__ = node.node.__dict__
+                old = node.node
+                replace_object_state(new, old)
                 node.node = new
-                # TODO: Other node types
-                if isinstance(node.node, Var) and node.node.type:
-                    node.node.type.accept(TypeReplaceVisitor(replacements))
-                    node.node.info = cast(TypeInfo, replacements.get(node.node.info,
-                                                                     node.node.info))
-            else:
-                # TODO: Other node types
-                if isinstance(node.node, Var) and node.node.type:
-                    node.node.type.accept(TypeReplaceVisitor(replacements))
-        override = node.type_override
-        if override:
-            override.accept(TypeReplaceVisitor(replacements))
+            if isinstance(node.node, Var):
+                # Handle them here just in case these aren't exposed through the AST.
+                # TODO: Is this necessary?
+                fixup_var(node.node, replacements)
+
+
+def fixup_var(node: Var, replacements: Dict[SymbolNode, SymbolNode]) -> None:
+    if node.type:
+        node.type.accept(TypeReplaceVisitor(replacements))
+    node.info = cast(TypeInfo, replacements.get(node.info, node.info))
